#[derive(Debug, Clone)]
pub enum Token {
    SEMI,
    NUMBER(String),
    IDENTIFIER(String),
    KEYWORD(Keyword),
}

#[derive(Debug, Clone)]
pub enum Keyword {
    RETURN,
}

enum TokenFinishedUnfinished {
    UNFINISHED(Option<Token>),
    FINISHED(Token),
}

use std::collections::HashMap;

use self::TokenFinishedUnfinished::*;

fn generate_keyword_map() -> HashMap<String, Keyword> {
    let mut keywords: HashMap<String, Keyword> = HashMap::new();

    use Keyword as K;
    keywords.insert("return".to_string(), K::RETURN);

    keywords
}

fn tokenise_current_character(
    current_character: char,
    current_token_type: Option<Token>,
) -> TokenFinishedUnfinished {
    let keywords = generate_keyword_map();

    match current_token_type {
        Some(Token::NUMBER(ref v)) => {
            let mut current_characters = v.clone();

            if current_character.is_digit(10) {
                current_characters.push(current_character);
                return UNFINISHED(Some(Token::NUMBER(current_characters)));
            }

            FINISHED(Token::NUMBER(current_characters))
        }
        Some(Token::IDENTIFIER(ref v)) => {
            let mut current_characters = v.clone();

            if current_character.is_alphanumeric() {
                current_characters.push(current_character);
                return UNFINISHED(Some(Token::IDENTIFIER(current_characters)));
            }

            match keywords.get(&current_characters) {
                Some(keyword) => FINISHED(Token::KEYWORD(keyword.clone())),
                None => FINISHED(Token::IDENTIFIER(current_characters)),
            }
        }
        Some(Token::SEMI) => FINISHED(Token::SEMI), // Needed because we're always going back a step after returning finished token
        None => {
            if current_character.is_digit(10) {
                return UNFINISHED(Some(Token::NUMBER(String::from(current_character))));
            }

            if current_character.is_alphabetic() {
                return UNFINISHED(Some(Token::IDENTIFIER(String::from(current_character))));
            }

            if current_character == ';' {
                return UNFINISHED(Some(Token::SEMI));
            }

            UNFINISHED(None)
        }
        Some(v) => FINISHED(v),
    }
}

pub fn tokenise(input: String) -> Vec<Token> {
    let mut tokens: Vec<Token> = vec![];

    let mut current_character: char;
    let mut current_token_type: Option<Token> = None;

    let mut i: usize = 0;
    while i < input.len() {
        current_character = input.as_bytes()[i] as char;

        match tokenise_current_character(current_character, current_token_type.clone()) {
            UNFINISHED(v) => {
                current_token_type = v.clone();
            }
            FINISHED(v) => {
                tokens.push(v);
                current_token_type = None;
                i -= 1;
            }
        }

        i += 1;
    }

    tokens
}
